@@ -4,23 +4,24 @@ from collections import defaultdict
 from decimal import Decimal
 
 from sql import Column
+from sql.conditionals import Coalesce
 from sql.operators import Concat
 from sql.aggregate import Count
 
 from trytond import backend
+from trytond.i18n import gettext
 from trytond.model import Workflow, ModelView, ModelSQL, fields
-from trytond.pyson import Eval, If
+from trytond.model.exceptions import AccessError
+from trytond.pyson import Eval, If, Id
 from trytond.pool import Pool, PoolMeta
 from trytond.transaction import Transaction
 from trytond.tools import grouped_slice, cursor_dict
 
 from trytond.modules.sale.stock import process_sale
+from trytond.modules.product import round_price
 from trytond.modules.purchase.stock import process_purchase
 
 
-__all__ = ['Configuration', 'ConfigurationSequence', 'ShipmentDrop', 'Move']
-
-
 class Configuration(metaclass=PoolMeta):
     __name__ = 'stock.configuration'
 
@@ -29,7 +30,9 @@ class Configuration(metaclass=PoolMeta):
             domain=[
                 ('company', 'in',
                     [Eval('context', {}).get('company', -1), None]),
-                ('code', '=', 'stock.shipment.drop'),
+                ('sequence_type', '=',
+                    Id('sale_supply_drop_shipment',
+                        'sequence_type_shipment_drop')),
                 ]))
 
     @classmethod
@@ -51,14 +54,15 @@ class ConfigurationSequence(metaclass=PoolMeta):
         'ir.sequence', "Drop Shipment Sequence", required=True,
         domain=[
             ('company', 'in', [Eval('company', -1), None]),
-            ('code', '=', 'stock.shipment.drop'),
+            ('sequence_type', '=',
+                Id('sale_supply_drop_shipment',
+                    'sequence_type_shipment_drop')),
             ],
         depends=['company'])
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        exist = TableHandler.table_exist(cls._table)
+        exist = backend.TableHandler.table_exist(cls._table)
         if exist:
             table = cls.__table_handler__(module_name)
             exist &= table.column_exist('shipment_drop_sequence')
@@ -89,8 +93,14 @@ class ConfigurationSequence(metaclass=PoolMeta):
 class ShipmentDrop(Workflow, ModelSQL, ModelView):
     "Drop Shipment"
     __name__ = 'stock.shipment.drop'
-
-    effective_date = fields.Date('Effective Date', readonly=True)
+    _rec_name = 'number'
+    effective_date = fields.Date(
+        "Effective Date",
+        states={
+            'readonly': Eval('state').in_(['cancelled', 'done']),
+            },
+        depends=['state'],
+        help="When the stock was actually sent.")
     planned_date = fields.Date('Planned Date', states={
             'readonly': Eval('state') != 'draft',
             }, depends=['state'])
@@ -98,10 +108,6 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         states={
             'readonly': Eval('state') != 'draft',
             },
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ],
         depends=['state'])
     reference = fields.Char('Reference', select=1,
         states={
@@ -113,7 +119,10 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
                     | Eval('supplier_moves', [0]))
                 & Eval('supplier')),
             },
-        depends=['state', 'supplier'])
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['state', 'supplier', 'company'])
     contact_address = fields.Many2One('party.address', 'Contact Address',
         states={
             'readonly': Eval('state') != 'draft',
@@ -126,7 +135,10 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
                     | Eval('customer_moves', [0]))
                 & Eval('customer')),
             },
-        depends=['state'])
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['state', 'company'])
     delivery_address = fields.Many2One('party.address', 'Delivery Address',
         required=True,
         states={
@@ -153,7 +165,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         'Supplier Moves',
         filter=[('to_location.type', '=', 'drop')],
         states={
-            'readonly': Eval('state').in_(['shipped', 'done', 'cancel']),
+            'readonly': Eval('state').in_(['shipped', 'done', 'cancelled']),
             },
         depends=['state', 'supplier'])
     customer_moves = fields.One2Many('stock.move', 'shipment',
@@ -163,13 +175,15 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
             'readonly': Eval('state') != 'shipped',
             },
         depends=['state', 'customer'])
-    code = fields.Char('Code', select=1, readonly=True)
+    number = fields.Char(
+        "Number", select=True, readonly=True,
+        help="The main identifier for the shipment.")
     state = fields.Selection([
             ('draft', 'Draft'),
             ('waiting', 'Waiting'),
             ('shipped', 'Shipped'),
             ('done', 'Done'),
-            ('cancel', 'Canceled'),
+            ('cancelled', 'Cancelled'),
             ], 'State', readonly=True)
 
     @classmethod
@@ -180,6 +194,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         PurchaseRequest = pool.get('purchase.request')
         SaleLine = pool.get('sale.line')
         Location = pool.get('stock.location')
+        table = cls.__table__()
         move = Move.__table__()
         purchase_line = PurchaseLine.__table__()
         purchase_request = PurchaseRequest.__table__()
@@ -187,6 +202,11 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         location = Location.__table__()
         cursor = Transaction().connection.cursor()
 
+        table_h = cls.__table_handler__(module_name)
+        # Migration from 5.8: rename code into number
+        if table_h.column_exist('code'):
+            table_h.column_rename('code', 'number')
+
         super(ShipmentDrop, cls).__register__(module_name)
 
         # Migration from 3.6
@@ -239,28 +259,37 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
                             drop_shipment_location, customer_location],
                         where=(move.id == move_id)))
 
+        # Migration from 5.6: rename state cancel to cancelled
+        cursor.execute(*table.update(
+                [table.state], ['cancelled'],
+                where=table.state == 'cancel'))
+
     @classmethod
     def __setup__(cls):
         super(ShipmentDrop, cls).__setup__()
+        cls._order = [
+            ('effective_date', 'ASC NULLS LAST'),
+            ('id', 'DESC'),
+            ]
         cls._transitions |= set((
                 ('draft', 'waiting'),
                 ('waiting', 'shipped'),
-                ('draft', 'cancel'),
-                ('waiting', 'cancel'),
+                ('draft', 'cancelled'),
+                ('waiting', 'cancelled'),
                 ('waiting', 'draft'),
-                ('cancel', 'draft'),
+                ('cancelled', 'draft'),
                 ('shipped', 'done'),
-                ('shipped', 'cancel'),
+                ('shipped', 'cancelled'),
                 ))
         cls._buttons.update({
                 'cancel': {
-                    'invisible': Eval('state').in_(['cancel', 'done']),
+                    'invisible': Eval('state').in_(['cancelled', 'done']),
                     'depends': ['state'],
                     },
                 'draft': {
-                    'invisible': ~Eval('state').in_(['cancel', 'draft',
+                    'invisible': ~Eval('state').in_(['cancelled', 'draft',
                             'waiting']),
-                    'icon': If(Eval('state') == 'cancel',
+                    'icon': If(Eval('state') == 'cancelled',
                         'tryton-undo', 'tryton-back'),
                     'depends': ['state'],
                     },
@@ -277,12 +306,11 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
                     'depends': ['state'],
                     },
                 })
-        cls._error_messages.update({
-                'reset_move': ('You cannot reset to draft move "%s" which was '
-                    'generated by a sale or a purchase.'),
-                'delete_cancel': ('Drop Shipment "%(shipment)s" must be '
-                    'cancelled before deletion.'),
-                })
+
+    @classmethod
+    def order_effective_date(cls, tables):
+        table, _ = tables[None]
+        return [Coalesce(table.effective_date, table.planned_date)]
 
     @staticmethod
     def default_state():
@@ -322,7 +350,8 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         for shipment in shipments:
             planned_date = shipment._get_move_planned_date()
             to_write.extend(([m for m in shipment.moves
-                        if m.state not in ('assigned', 'done', 'cancel')], {
+                        if m.state not in ('assigned', 'done', 'cancelled')
+                        ], {
                         'planned_date': planned_date,
                         }))
         if to_write:
@@ -331,13 +360,15 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Config = pool.get('stock.configuration')
 
         vlist = [x.copy() for x in vlist]
         config = Config(1)
+        default_company = cls.default_company()
         for values in vlist:
-            values['code'] = Sequence.get_id(config.shipment_drop_sequence)
+            values['number'] = config.get_multivalue(
+                'shipment_drop_sequence',
+                company=values.get('company', default_company)).get()
         shipments = super(ShipmentDrop, cls).create(vlist)
         cls._set_move_planned_date(shipments)
         return shipments
@@ -363,8 +394,10 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
 
         cls.cancel(shipments)
         for shipment in shipments:
-            if shipment.state != 'cancel':
-                cls.raise_user_error('delete_cancel', {
+            if shipment.state != 'cancelled':
+                raise AccessError(
+                    gettext('sale_supply_drop_shipment'
+                        '.msg_drop_shipment_delete_cancel') % {
                         'shipment': shipment.rec_name,
                         })
         Move.delete([m for s in shipments for m in s.supplier_moves])
@@ -372,7 +405,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     @process_sale('customer_moves')
     @process_purchase('supplier_moves')
     def cancel(cls, shipments):
@@ -393,9 +426,11 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         SaleLine = pool.get('sale.line')
         for shipment in shipments:
             for move in shipment.moves:
-                if (move.state == 'cancel'
+                if (move.state == 'cancelled'
                         and isinstance(move.origin, (PurchaseLine, SaleLine))):
-                    cls.raise_user_error('reset_move', (move.rec_name,))
+                    raise AccessError(
+                        gettext('sale_supply_drop_shipment.msg_reset_move',
+                            move=move.rec_name))
         Move.draft([m for s in shipments for m in s.moves
                 if m.state != 'staging'])
 
@@ -406,26 +441,24 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         Move = pool.get('stock.move')
 
         to_save = []
-        cost_exp = Decimal(str(10.0 ** -Move.cost_price.digits[1]))
         for shipment in shipments:
-            product_qty = defaultdict(lambda: 0)
-            product_cost = defaultdict(lambda: 0)
+            product_qty = defaultdict(int)
+            product_cost = defaultdict(int)
 
             for c_move in shipment.customer_moves:
-                if c_move.state == 'cancel':
+                if c_move.state == 'cancelled':
                     continue
                 product_qty[c_move.product] += UoM.compute_qty(
                     c_move.uom, c_move.quantity, c_move.product.default_uom,
                     round=False)
 
-            s_product_qty = defaultdict(lambda: 0)
+            s_product_qty = defaultdict(int)
             for s_move in shipment.supplier_moves:
-                if s_move.state == 'cancel':
+                if s_move.state == 'cancelled':
                     continue
-                if s_move.cost_price:
-                    internal_quantity = Decimal(str(s_move.internal_quantity))
-                    product_cost[s_move.product] += (
-                        s_move.unit_price * internal_quantity)
+                internal_quantity = Decimal(str(s_move.internal_quantity))
+                product_cost[s_move.product] += (
+                    s_move.get_cost_price() * internal_quantity)
 
                 quantity = UoM.compute_qty(
                     s_move.uom, s_move.quantity, s_move.product.default_uom,
@@ -458,9 +491,9 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
             for product, cost in product_cost.items():
                 qty = Decimal(str(s_product_qty[product]))
                 if qty:
-                    product_cost[product] = (cost / qty).quantize(cost_exp)
+                    product_cost[product] = round_price(cost / qty)
             for c_move in list(shipment.customer_moves) + to_save:
-                if c_move.id is not None and c_move.state == 'cancel':
+                if c_move.id is not None and c_move.state == 'cancelled':
                     continue
                 c_move.cost_price = product_cost[c_move.product]
                 if c_move.id is None:
@@ -492,8 +525,6 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
             shipment=self,
             planned_date=self.planned_date,
             company=move.company,
-            currency=move.company.currency,
-            unit_price=move.unit_price,
             )
 
     @classmethod
@@ -514,9 +545,11 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         pline2requests = defaultdict(list)
         for request in requests:
             pline2requests[request.purchase_line].append(request)
-        sale_lines = SaleLine.search([
-                ('purchase_request', 'in', [r.id for r in requests]),
-                ])
+        sale_lines = []
+        for sub_requests in grouped_slice([r.id for r in requests]):
+            sale_lines += SaleLine.search([
+                    ('purchase_request', 'in', list(sub_requests)),
+                    ])
         request2slines = defaultdict(list)
         for sale_line in sale_lines:
             request2slines[sale_line.purchase_request].append(sale_line)
@@ -529,7 +562,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
                 for request in pline2requests[move.origin]:
                     for sale_line in request2slines[request]:
                         for move in sale_line.moves:
-                            if (move.state not in ('cancel', 'done')
+                            if (move.state not in ('cancelled', 'done')
                                     and not move.shipment
                                     and move.from_location.type == 'drop'):
                                 move.shipment = shipment
@@ -556,7 +589,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         Move = pool.get('stock.move')
         Date = pool.get('ir.date')
         Move.do([m for s in shipments for m in s.customer_moves])
-        cls.write(shipments, {
+        cls.write([s for s in shipments if not s.effective_date], {
                 'effective_date': Date.today(),
                 })
 
@@ -564,8 +597,13 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
 class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
 
-    customer_drop = fields.Function(fields.Many2One('party.party',
-            'Drop Customer'), 'get_customer_drop',
+    customer_drop = fields.Function(fields.Many2One(
+            'party.party', "Drop Customer",
+            context={
+                'company': Eval('company', -1),
+                },
+            depends=['company']),
+        'get_customer_drop',
         searcher='search_customer_drop')
 
     @classmethod
@@ -582,7 +620,8 @@ class Move(metaclass=PoolMeta):
         if isinstance(self.origin, SaleLine):
             return self.origin.sale.party.id
         elif isinstance(self.origin, PurchaseLine):
-            return self.origin.purchase.customer.id
+            if self.origin.purchase.customer:
+                return self.origin.purchase.customer.id
 
     @classmethod
     def search_customer_drop(cls, name, clause):
